@@ -4,7 +4,7 @@
  * in here.
  *
  * ====================================================================
- * Copyright (c) 2000-2006 CollabNet.  All rights reserved.
+ * Copyright (c) 2000-2007 CollabNet.  All rights reserved.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution.  The terms
@@ -28,22 +28,28 @@
 #include <ctype.h>
 #include <assert.h>
 
+#include <apr_env.h>
 #include <apr_errno.h>
+#include <apr_file_info.h>
 #include <apr_strings.h>
 #include <apr_tables.h>
 #include <apr_general.h>
 #include <apr_lib.h>
 
+#include "svn_pools.h"
+#include "svn_error.h"
+#include "svn_ctype.h"
 #include "svn_client.h"
 #include "svn_cmdline.h"
 #include "svn_string.h"
 #include "svn_path.h"
-#include "svn_error.h"
+#include "svn_hash.h"
 #include "svn_io.h"
 #include "svn_utf.h"
 #include "svn_subst.h"
 #include "svn_config.h"
 #include "svn_xml.h"
+#include "svn_time.h"
 #include "svn_private_config.h"
 #include "cl.h"
 
@@ -54,36 +60,207 @@ svn_error_t *
 svn_cl__print_commit_info(svn_commit_info_t *commit_info,
                           apr_pool_t *pool)
 {
-  if (commit_info) 
-    {
-      if (SVN_IS_VALID_REVNUM(commit_info->revision))
-        SVN_ERR(svn_cmdline_printf(pool, _("\nCommitted revision %ld.\n"),
-                                   commit_info->revision));
-     
-      /* Writing to stdout, as there maybe systems that consider the
-       * presence of stderr as an indication of commit failure.
-       * OTOH, this is only of informational nature to the user as
-       * the commit has succeeded. */
-      if (commit_info->post_commit_err)
-        SVN_ERR(svn_cmdline_printf(pool, _("\nWarning: %s\n"),
-                                   commit_info->post_commit_err));
-    }
+  if (SVN_IS_VALID_REVNUM(commit_info->revision))
+    SVN_ERR(svn_cmdline_printf(pool, _("\nCommitted revision %ld.\n"),
+                               commit_info->revision));
+
+  /* Writing to stdout, as there maybe systems that consider the
+   * presence of stderr as an indication of commit failure.
+   * OTOH, this is only of informational nature to the user as
+   * the commit has succeeded. */
+  if (commit_info->post_commit_err)
+    SVN_ERR(svn_cmdline_printf(pool, _("\nWarning: %s\n"),
+                               commit_info->post_commit_err));
 
   return SVN_NO_ERROR;
 }
 
 
+/* Helper for the next two functions.  Set *EDITOR to some path to an
+   editor binary.  Sources to search include: the EDITOR_CMD argument
+   (if not NULL), $SVN_EDITOR, the runtime CONFIG variable (if CONFIG
+   is not NULL), $VISUAL, $EDITOR.  Return
+   SVN_ERR_CL_NO_EXTERNAL_EDITOR if no binary can be found. */
+static svn_error_t *
+find_editor_binary(const char **editor,
+                   const char *editor_cmd,
+                   apr_hash_t *config)
+{
+  const char *e;
+  struct svn_config_t *cfg;
+
+  /* Use the editor specified on the command line via --editor-cmd, if any. */
+  e = editor_cmd;
+
+  /* Otherwise look for the Subversion-specific environment variable. */
+  if (! e)
+    e = getenv("SVN_EDITOR");
+
+  /* If not found then fall back on the config file. */
+  if (! e)
+    {
+      cfg = config ? apr_hash_get(config, SVN_CONFIG_CATEGORY_CONFIG,
+                                  APR_HASH_KEY_STRING) : NULL;
+      svn_config_get(cfg, &e, SVN_CONFIG_SECTION_HELPERS,
+                     SVN_CONFIG_OPTION_EDITOR_CMD, NULL);
+    }
+
+  /* If not found yet then try general purpose environment variables. */
+  if (! e)
+    e = getenv("VISUAL");
+  if (! e)
+    e = getenv("EDITOR");
+
+#ifdef SVN_CLIENT_EDITOR
+  /* If still not found then fall back on the hard-coded default. */
+  if (! e)
+    e = SVN_CLIENT_EDITOR;
+#endif
+
+  /* Error if there is no editor specified */
+  if (e)
+    {
+      const char *c;
+
+      for (c = e; *c; c++)
+        if (!svn_ctype_isspace(*c))
+          break;
+
+      if (! *c)
+        return svn_error_create
+          (SVN_ERR_CL_NO_EXTERNAL_EDITOR, NULL,
+           _("The EDITOR, SVN_EDITOR or VISUAL environment variable or "
+             "'editor-cmd' run-time configuration option is empty or "
+             "consists solely of whitespace. Expected a shell command."));
+    }
+  else
+    return svn_error_create
+      (SVN_ERR_CL_NO_EXTERNAL_EDITOR, NULL,
+       _("None of the environment variables SVN_EDITOR, VISUAL or EDITOR are "
+         "set, and no 'editor-cmd' run-time configuration option was found"));
+
+  *editor = e;
+  return SVN_NO_ERROR;
+}
+
+
+/* Use the visual editor to edit files. This requires that the file name itself
+   be shell-safe, although the path to reach that file need not be shell-safe.
+ */
 svn_error_t *
-svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
-                        const char **tmpfile_left /* UTF-8! */,
-                        const char *editor_cmd,
-                        const char *base_dir /* UTF-8! */,
-                        const svn_string_t *contents /* UTF-8! */,
-                        const char *prefix,
-                        apr_hash_t *config,
-                        svn_boolean_t as_text,
-                        const char *encoding,
-                        apr_pool_t *pool)
+svn_cl__edit_file_externally(const char *path,
+                             const char *editor_cmd,
+                             apr_hash_t *config,
+                             apr_pool_t *pool)
+{
+  const char *editor, *cmd, *base_dir, *file_name, *base_dir_apr;
+  char *old_cwd;
+  int sys_err;
+  apr_status_t apr_err;
+
+  svn_path_split(path, &base_dir, &file_name, pool);
+
+  SVN_ERR(find_editor_binary(&editor, editor_cmd, config));
+
+  apr_err = apr_filepath_get(&old_cwd, APR_FILEPATH_NATIVE, pool);
+  if (apr_err)
+    return svn_error_wrap_apr(apr_err, _("Can't get working directory"));
+
+  /* APR doesn't like "" directories */
+  if (base_dir[0] == '\0')
+    base_dir_apr = ".";
+  else
+    SVN_ERR(svn_path_cstring_from_utf8(&base_dir_apr, base_dir, pool));
+
+  apr_err = apr_filepath_set(base_dir_apr, pool);
+  if (apr_err)
+    return svn_error_wrap_apr
+      (apr_err, _("Can't change working directory to '%s'"), base_dir);
+
+  cmd = apr_psprintf(pool, "%s %s", editor, file_name);
+  sys_err = system(cmd);
+
+  apr_err = apr_filepath_set(old_cwd, pool);
+  if (apr_err)
+    svn_handle_error2(svn_error_wrap_apr
+                      (apr_err, _("Can't restore working directory")),
+                      stderr, TRUE /* fatal */, "svn: ");
+
+  if (sys_err)
+    /* Extracting any meaning from sys_err is platform specific, so just
+       use the raw value. */
+    return svn_error_createf(SVN_ERR_EXTERNAL_PROGRAM, NULL,
+                             _("system('%s') returned %d"), cmd, sys_err);
+
+  return SVN_NO_ERROR;
+}
+
+svn_error_t *
+svn_cl__merge_file_externally(const char *base_path,
+                              const char *their_path,
+                              const char *my_path,
+                              const char *merged_path,
+                              apr_hash_t *config,
+                              apr_pool_t *pool)
+{
+  char *merge_tool;
+  /* Error if there is no editor specified */
+  if (apr_env_get(&merge_tool, "SVN_MERGE", pool) != APR_SUCCESS)
+    {
+      struct svn_config_t *cfg;
+      merge_tool = NULL;
+      cfg = config ? apr_hash_get(config, SVN_CONFIG_CATEGORY_CONFIG,
+                                  APR_HASH_KEY_STRING) : NULL;
+      /* apr_env_get wants char **, this wants const char ** */
+      svn_config_get(cfg, (const char **)&merge_tool,
+                     SVN_CONFIG_SECTION_HELPERS,
+                     SVN_CONFIG_OPTION_MERGE_TOOL_CMD, NULL);
+    }
+
+  if (merge_tool)
+    {
+      const char *c;
+
+      for (c = merge_tool; *c; c++)
+        if (!svn_ctype_isspace(*c))
+          break;
+
+      if (! *c)
+        return svn_error_create
+          (SVN_ERR_CL_NO_EXTERNAL_MERGE_TOOL, NULL,
+           _("The SVN_MERGE environment variable is empty or "
+             "consists solely of whitespace. Expected a shell command.\n"));
+    }
+  else
+      return svn_error_create
+        (SVN_ERR_CL_NO_EXTERNAL_MERGE_TOOL, NULL,
+         _("The environment variable SVN_MERGE and the merge-tool-cmd run-time "
+           "configuration option were not set.\n"));
+
+  {
+    const char *arguments[] = { merge_tool, base_path, their_path,
+                                my_path, merged_path, NULL};
+    char *cwd;
+    apr_status_t status = apr_filepath_get(&cwd, APR_FILEPATH_NATIVE, pool);
+    if (status != 0)
+      return svn_error_wrap_apr(status, NULL);
+    return svn_io_run_cmd(svn_path_internal_style(cwd, pool), merge_tool,
+                          arguments, NULL, NULL, TRUE, NULL, NULL, NULL,
+                          pool);
+  }
+}
+
+svn_error_t *
+svn_cl__edit_string_externally(svn_string_t **edited_contents /* UTF-8! */,
+                               const char **tmpfile_left /* UTF-8! */,
+                               const char *editor_cmd,
+                               const char *base_dir /* UTF-8! */,
+                               const svn_string_t *contents /* UTF-8! */,
+                               const char *prefix,
+                               apr_hash_t *config,
+                               svn_boolean_t as_text,
+                               const char *encoding,
+                               apr_pool_t *pool)
 {
   const char *editor;
   const char *cmd;
@@ -99,42 +276,8 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
   char *old_cwd;
   int sys_err;
   svn_boolean_t remove_file = TRUE;
-  struct svn_config_t *cfg;
 
-  /* Use the editor specified on the command line via --editor-cmd, if any. */
-  editor = editor_cmd;
-
-  /* Otherwise look for the Subversion-specific environment variable. */
-  if (! editor)
-    editor = getenv("SVN_EDITOR");
-
-  /* If not found then fall back on the config file. */
-  if (! editor)
-    {
-      cfg = config ? apr_hash_get(config, SVN_CONFIG_CATEGORY_CONFIG, 
-                                  APR_HASH_KEY_STRING) : NULL;
-      svn_config_get(cfg, &editor, SVN_CONFIG_SECTION_HELPERS, 
-                     SVN_CONFIG_OPTION_EDITOR_CMD, NULL);
-    }
-
-  /* If not found yet then try general purpose environment variables. */
-  if (! editor)
-    editor = getenv("VISUAL");
-  if (! editor)
-    editor = getenv("EDITOR");
-
-#ifdef SVN_CLIENT_EDITOR
-  /* If still not found then fall back on the hard-coded default. */
-  if (! editor)
-    editor = SVN_CLIENT_EDITOR;
-#endif
-
-  /* Abort if there is no editor specified */
-  if (! editor)
-    return svn_error_create 
-      (SVN_ERR_CL_NO_EXTERNAL_EDITOR, NULL,
-       _("None of the environment variables SVN_EDITOR, VISUAL or EDITOR is "
-         "set, and no 'editor-cmd' run-time configuration option was found"));
+  SVN_ERR(find_editor_binary(&editor, editor_cmd, config));
 
   /* Convert file contents from UTF-8/LF if desired. */
   if (as_text)
@@ -180,7 +323,7 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
   err = svn_io_open_unique_file2(&tmp_file, &tmpfile_name,
                                  prefix, ".tmp", svn_io_file_del_none, pool);
 
-  if (err && APR_STATUS_IS_EACCES(err->apr_err))
+  if (err && (APR_STATUS_IS_EACCES(err->apr_err) || err->apr_err == EROFS))
     {
       const char *temp_dir_apr;
 
@@ -214,11 +357,11 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
   apr_err2 = apr_file_close(tmp_file);
   if (! apr_err)
     apr_err = apr_err2;
-  
+
   /* Make sure the whole CONTENTS were written, else return an error. */
   if (apr_err)
     {
-      err = svn_error_wrap_apr(apr_err, _("Can't write to '%s'"), 
+      err = svn_error_wrap_apr(apr_err, _("Can't write to '%s'"),
                                tmpfile_name);
       goto cleanup;
     }
@@ -229,6 +372,21 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
 
   /* Get information about the temporary file before the user has
      been allowed to edit its contents. */
+  apr_err = apr_stat(&finfo_before, tmpfile_apr,
+                     APR_FINFO_MTIME, pool);
+  if (apr_err)
+    {
+      err = svn_error_wrap_apr(apr_err, _("Can't stat '%s'"), tmpfile_name);
+      goto cleanup;
+    }
+
+  /* Backdate the file a little bit in case the editor is very fast
+     and doesn't change the size.  (Use two seconds, since some
+     filesystems have coarse granularity.)  It's OK if this call
+     fails, so we don't check its return value.*/
+  apr_file_mtime_set(tmpfile_apr, finfo_before.mtime - 2000, pool);
+
+  /* Stat it again to get the mtime we actually set. */
   apr_err = apr_stat(&finfo_before, tmpfile_apr,
                      APR_FINFO_MTIME | APR_FINFO_SIZE, pool);
   if (apr_err)
@@ -251,7 +409,7 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
                                _("system('%s') returned %d"), cmd, sys_err);
       goto cleanup;
     }
-  
+
   /* Get information about the temporary file after the assumed editing. */
   apr_err = apr_stat(&finfo_after, tmpfile_apr,
                      APR_FINFO_MTIME | APR_FINFO_SIZE, pool);
@@ -286,7 +444,12 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
           err = svn_subst_translate_string(edited_contents, *edited_contents,
                                            encoding, pool);
           if (err)
-            goto cleanup;
+            {
+              err = svn_error_quick_wrap
+                (err,
+                 _("Error normalizing edited contents to internal format"));
+              goto cleanup;
+            }
         }
     }
   else
@@ -323,6 +486,8 @@ svn_cl__edit_externally(svn_string_t **edited_contents /* UTF-8! */,
 }
 
 
+/* A svn_client_ctx_t's log_msg_baton3, for use with
+   svn_cl__make_log_msg_baton(). */
 struct log_msg_baton
 {
   const char *editor_cmd;  /* editor specified via --editor-cmd, else NULL */
@@ -346,7 +511,7 @@ svn_cl__make_log_msg_baton(void **baton,
 {
   struct log_msg_baton *lmb = apr_palloc(pool, sizeof(*lmb));
 
-  if (opt_state->filedata) 
+  if (opt_state->filedata)
     {
       if (strlen(opt_state->filedata->data) < opt_state->filedata->len)
         {
@@ -358,7 +523,7 @@ svn_cl__make_log_msg_baton(void **baton,
                                   _("Log message contains a zero byte"));
         }
       lmb->message = opt_state->filedata->data;
-    }      
+    }
   else
     {
       lmb->message = opt_state->message;
@@ -371,7 +536,7 @@ svn_cl__make_log_msg_baton(void **baton,
     }
   else if (config)
     {
-      svn_config_t *cfg = apr_hash_get(config, SVN_CONFIG_CATEGORY_CONFIG, 
+      svn_config_t *cfg = apr_hash_get(config, SVN_CONFIG_CATEGORY_CONFIG,
                                        APR_HASH_KEY_STRING);
       svn_config_get(cfg, &(lmb->message_encoding),
                      SVN_CONFIG_SECTION_MISCELLANY,
@@ -469,7 +634,6 @@ truncate_buffer_at_prefix(apr_size_t *new_len,
 
 #define EDITOR_EOF_PREFIX  _("--This line, and those below, will be ignored--")
 
-/* This function is of type svn_client_get_commit_log2_t. */
 svn_error_t *
 svn_cl__get_log_message(const char **log_msg,
                         const char **tmp_file,
@@ -491,8 +655,9 @@ svn_cl__get_log_message(const char **log_msg,
     {
       svn_string_t *log_msg_string = svn_string_create(lmb->message, pool);
 
-      SVN_ERR(svn_subst_translate_string(&log_msg_string, log_msg_string,
-                                         lmb->message_encoding, pool));
+      SVN_ERR_W(svn_subst_translate_string(&log_msg_string, log_msg_string,
+                                           lmb->message_encoding, pool),
+                _("Error normalizing log message to internal format"));
 
       *log_msg = log_msg_string->data;
 
@@ -530,8 +695,8 @@ svn_cl__get_log_message(const char **log_msg,
 
       for (i = 0; i < commit_items->nelts; i++)
         {
-          svn_client_commit_item2_t *item
-            = APR_ARRAY_IDX(commit_items, i, svn_client_commit_item2_t *);
+          svn_client_commit_item3_t *item
+            = APR_ARRAY_IDX(commit_items, i, svn_client_commit_item3_t *);
           const char *path = item->path;
           char text_mod = '_', prop_mod = ' ', unlock = ' ';
 
@@ -564,10 +729,14 @@ svn_cl__get_log_message(const char **log_msg,
               && item->state_flags & SVN_CLIENT_COMMIT_ITEM_LOCK_TOKEN)
             unlock = 'U';
 
-          svn_stringbuf_appendbytes(tmp_message, &text_mod, 1); 
-          svn_stringbuf_appendbytes(tmp_message, &prop_mod, 1); 
-          svn_stringbuf_appendbytes(tmp_message, &unlock, 1); 
-          svn_stringbuf_appendcstr(tmp_message, "  ");
+          svn_stringbuf_appendbytes(tmp_message, &text_mod, 1);
+          svn_stringbuf_appendbytes(tmp_message, &prop_mod, 1);
+          svn_stringbuf_appendbytes(tmp_message, &unlock, 1);
+          if (item->state_flags & SVN_CLIENT_COMMIT_ITEM_IS_COPY)
+            /* History included via copy/move. */
+            svn_stringbuf_appendcstr(tmp_message, "+ ");
+          else
+            svn_stringbuf_appendcstr(tmp_message, "  ");
           svn_stringbuf_appendcstr(tmp_message, path);
           svn_stringbuf_appendcstr(tmp_message, APR_EOL_STR);
         }
@@ -578,12 +747,12 @@ svn_cl__get_log_message(const char **log_msg,
       /* Use the external edit to get a log message. */
       if (! lmb->non_interactive)
         {
-          err = svn_cl__edit_externally(&msg_string, &lmb->tmpfile_left,
-                                        lmb->editor_cmd, lmb->base_dir,
-                                        msg_string, "svn-commit",
-                                        lmb->config, TRUE,
-                                        lmb->message_encoding,
-                                        pool);
+          err = svn_cl__edit_string_externally(&msg_string, &lmb->tmpfile_left,
+                                               lmb->editor_cmd, lmb->base_dir,
+                                               msg_string, "svn-commit",
+                                               lmb->config, TRUE,
+                                               lmb->message_encoding,
+                                               pool);
         }
       else /* non_interactive flag says we can't pop up an editor, so error */
         {
@@ -594,14 +763,14 @@ svn_cl__get_log_message(const char **log_msg,
         }
 
       /* Dup the tmpfile path into its baton's pool. */
-      *tmp_file = lmb->tmpfile_left = apr_pstrdup(lmb->pool, 
+      *tmp_file = lmb->tmpfile_left = apr_pstrdup(lmb->pool,
                                                   lmb->tmpfile_left);
 
       /* If the edit returned an error, handle it. */
       if (err)
         {
           if (err->apr_err == SVN_ERR_CL_NO_EXTERNAL_EDITOR)
-            err = svn_error_quick_wrap 
+            err = svn_error_quick_wrap
               (err, _("Could not use external editor to fetch log message; "
                       "consider setting the $SVN_EDITOR environment variable "
                       "or using the --message (-m) or --file (-F) options"));
@@ -613,7 +782,7 @@ svn_cl__get_log_message(const char **log_msg,
 
       /* Strip the prefix from the buffer. */
       if (message)
-        truncate_buffer_at_prefix(&message->len, message->data, 
+        truncate_buffer_at_prefix(&message->len, message->data,
                                   EDITOR_EOF_PREFIX);
 
       if (message)
@@ -639,7 +808,7 @@ svn_cl__get_log_message(const char **log_msg,
           SVN_ERR(svn_cmdline_prompt_user
                   (&reply,
                    _("\nLog message unchanged or not specified\n"
-                     "a)bort, c)ontinue, e)dit\n"), pool));
+                     "(a)bort, (c)ontinue, (e)dit :\n"), pool));
           if (reply)
             {
               char letter = apr_tolower(reply[0]);
@@ -657,7 +826,7 @@ svn_cl__get_log_message(const char **log_msg,
               /* If the user chooses to continue, we make an empty
                  message, which will cause us to exit the loop.  We
                  also cleanup the temporary file. */
-              if ('c' == letter) 
+              if ('c' == letter)
                 {
                   SVN_ERR(svn_io_remove_file(lmb->tmpfile_left, pool));
                   *tmp_file = lmb->tmpfile_left = NULL;
@@ -669,7 +838,7 @@ svn_cl__get_log_message(const char **log_msg,
             }
         }
     }
-  
+
   *log_msg = message ? message->data : NULL;
   return SVN_NO_ERROR;
 }
@@ -774,6 +943,59 @@ svn_cl__xml_tagged_cdata(svn_stringbuf_t **sb,
 }
 
 
+void
+svn_cl__print_xml_commit(svn_stringbuf_t **sb,
+                         svn_revnum_t revision,
+                         const char *author,
+                         const char *date,
+                         apr_pool_t *pool)
+{
+  /* "<commit ...>" */
+  svn_xml_make_open_tag(sb, pool, svn_xml_normal, "commit",
+                        "revision",
+                        apr_psprintf(pool, "%ld", revision), NULL);
+
+  /* "<author>xx</author>" */
+  if (author)
+    svn_cl__xml_tagged_cdata(sb, pool, "author", author);
+
+  /* "<date>xx</date>" */
+  if (date)
+    svn_cl__xml_tagged_cdata(sb, pool, "date", date);
+
+  /* "</commit>" */
+  svn_xml_make_close_tag(sb, pool, "commit");
+}
+
+
+svn_error_t *
+svn_cl__xml_print_header(const char *tagname,
+                         apr_pool_t *pool)
+{
+  svn_stringbuf_t *sb = svn_stringbuf_create("", pool);
+
+  /* <?xml version="1.0"?> */
+  svn_xml_make_header(&sb, pool);
+
+  /* "<TAGNAME>" */
+  svn_xml_make_open_tag(&sb, pool, svn_xml_normal, tagname, NULL);
+
+  return svn_cl__error_checked_fputs(sb->data, stdout);
+}
+
+
+svn_error_t *
+svn_cl__xml_print_footer(const char *tagname,
+                         apr_pool_t *pool)
+{
+  svn_stringbuf_t *sb = svn_stringbuf_create("", pool);
+
+  /* "</TAGNAME>" */
+  svn_xml_make_close_tag(&sb, pool, tagname);
+  return svn_cl__error_checked_fputs(sb->data, stdout);
+}
+
+
 const char *
 svn_cl__node_kind_str(svn_node_kind_t kind)
 {
@@ -787,3 +1009,111 @@ svn_cl__node_kind_str(svn_node_kind_t kind)
       return "";
     }
 }
+
+
+svn_error_t *
+svn_cl__args_to_target_array_print_reserved(apr_array_header_t **targets,
+                                            apr_getopt_t *os,
+                                            apr_array_header_t *known_targets,
+                                            apr_pool_t *pool)
+{
+  svn_error_t *error = svn_opt_args_to_target_array3(targets, os,
+                                                     known_targets, pool);
+  if (error)
+    {
+      if (error->apr_err ==  SVN_ERR_RESERVED_FILENAME_SPECIFIED)
+        {
+          svn_handle_error2(error, stderr, FALSE, "svn: Skipping argument: ");
+          svn_error_clear(error);
+        }
+      else
+        return error;
+    }
+  return SVN_NO_ERROR;
+}
+
+
+/* Helper for svn_cl__get_changelist(); implements
+   svn_changelist_receiver_t. */
+static svn_error_t *
+changelist_receiver(void *baton,
+                    const char *path,
+                    const char *changelist,
+                    apr_pool_t *pool)
+{
+  /* No need to check CHANGELIST; our caller only asked about one of them. */
+  apr_array_header_t *paths = baton;
+  APR_ARRAY_PUSH(paths, const char *) = apr_pstrdup(paths->pool, path);
+  return SVN_NO_ERROR;
+}
+
+
+svn_error_t *
+svn_cl__changelist_paths(apr_array_header_t **paths,
+                         const apr_array_header_t *changelists,
+                         const apr_array_header_t *targets,
+                         svn_depth_t depth,
+                         svn_client_ctx_t *ctx,
+                         apr_pool_t *pool)
+{
+  apr_array_header_t *found;
+  apr_pool_t *subpool = svn_pool_create(pool);
+  apr_hash_t *paths_hash;
+  int i;
+
+  if (! (changelists && changelists->nelts))
+    {
+      *paths = (apr_array_header_t *)targets;
+      return SVN_NO_ERROR;
+    }
+
+  found = apr_array_make(pool, 8, sizeof(const char *));
+  for (i = 0; i < targets->nelts; i++)
+    {
+      const char *target = APR_ARRAY_IDX(targets, i, const char *);
+      svn_pool_clear(subpool);
+      SVN_ERR(svn_client_get_changelists(target, changelists, depth,
+                                         changelist_receiver, (void *)found,
+                                         ctx, subpool));
+    }
+  svn_pool_destroy(subpool);
+
+  SVN_ERR(svn_hash_from_cstring_keys(&paths_hash, found, pool));
+  return svn_hash_keys(paths, paths_hash, pool);
+}
+
+svn_cl__show_revs_t
+svn_cl__show_revs_from_word(const char *word)
+{
+  if (strcmp(word, SVN_CL__SHOW_REVS_MERGED) == 0)
+    return svn_cl__show_revs_merged;
+  if (strcmp(word, SVN_CL__SHOW_REVS_ELIGIBLE) == 0)
+    return svn_cl__show_revs_eligible;
+  /* word is an invalid flavor. */
+  return svn_cl__show_revs_invalid;
+}
+
+
+svn_error_t *
+svn_cl__time_cstring_to_human_cstring(const char **human_cstring,
+                                      const char *data,
+                                      apr_pool_t *pool)
+{
+  svn_error_t *err;
+  apr_time_t when;
+
+  err = svn_time_from_cstring(&when, data, pool);
+  if (err && err->apr_err == SVN_ERR_BAD_DATE)
+    {
+      svn_error_clear(err);
+
+      *human_cstring = _("(invalid date)");
+      return SVN_NO_ERROR;
+    }
+  else if (err)
+    return err;
+
+  *human_cstring = svn_time_to_human_cstring(when, pool);
+
+  return SVN_NO_ERROR;
+}
